"""
paradigm.py
this script is response for providing a base class called Paradigm.
"""


class Paradigm:
    """
    it's base for many class related to math.
    """
    def __init__(self):
        pass

    def basic_data_type(self):
        pass

    def formula(self):
        pass
